/*
 * resume.c - Functions for waking devices up.
 *
 * Copyright (c) 2003 Patrick Mochel
 * Copyright (c) 2003 Open Source Development Labs
 *
 * This file is released under the GPLv2
 *
 */

#include <linux/device.h>
#include "power.h"

extern int sysdev_resume(void);


/**
 *	resume_device - Restore state for one device.
 *	@dev:	Device.
 *
 */

int resume_device(struct device * dev)
{
	if (dev->bus && dev->bus->resume)
		return dev->bus->resume(dev);
	return 0;
}



void dpm_resume_tree(struct partial_device_tree * tree)
{
	while(!list_empty(&tree->dpm_off)) {
		struct list_head * entry = tree->dpm_off.next;
		struct device * dev = to_device(entry);
		list_del_init(entry);

		if (!dev->power.prev_state)
			resume_device(dev);

		list_add_tail(entry, &tree->dpm_active);
		dev->current_list = DEVICE_LIST_DPM_ACTIVE;
	}
}

void dpm_resume(void)
{
	dpm_resume_tree(&default_device_tree);
}

/**
 *	device_resume - Restore state of each device in system.
 *
 *	Walk the dpm_off list, remove each entry, resume the device,
 *	then add it to the dpm_active list.
 */

void device_resume(void)
{
	down(&dpm_sem);
	dpm_resume();
	up(&dpm_sem);
}

EXPORT_SYMBOL(device_resume);

void device_resume_tree(struct partial_device_tree * tree)
{
	down(&dpm_sem);
	dpm_resume_tree(tree);
	up(&dpm_sem);
}

EXPORT_SYMBOL(device_resume_tree);

/**
 *	device_power_up_irq - Power on some devices.
 *
 *	Walk the dpm_off_irq list and power each device up. This
 *	is used for devices that required they be powered down with
 *	interrupts disabled. As devices are powered on, they are moved to
 *	the dpm_suspended list.
 *
 *	Interrupts must be disabled when calling this.
 */

void dpm_power_up_tree(struct partial_device_tree * tree)
{
	while(!list_empty(&tree->dpm_off_irq)) {
		struct list_head * entry = tree->dpm_off_irq.next;
		struct device * dev = to_device(entry);
		list_del_init(entry);
		resume_device(dev);
		list_add_tail(entry, &tree->dpm_active);
		dev->current_list = DEVICE_LIST_DPM_ACTIVE;
	}
}
EXPORT_SYMBOL(dpm_power_up_tree);

void dpm_power_up(void)
{
	dpm_power_up_tree(&default_device_tree);
}

/**
 *	device_pm_power_up - Turn on all devices that need special attention.
 *
 *	Power on system devices then devices that required we shut them down
 *	with interrupts disabled.
 *	Called with interrupts disabled.
 */

void device_power_up(void)
{
	sysdev_resume();
	dpm_power_up();
}

#if 0

/**
 *
 * 	pci_find_class_storage
 *
 *	Find a PCI storage device.
 *	Based upon pci_find_class, but less strict.
 */

static struct pci_dev *
pci_find_class_storage(unsigned int class, const struct pci_dev *from)
{
	struct list_head *n;
	struct pci_dev *dev;

	spin_lock(&pci_bus_lock);
	n = from ? from->global_list.next : pci_devices.next;

	while (n && (n != &pci_devices)) {
		dev = pci_dev_g(n);
		if (((dev->class & 0xff00) >> 16) == class)
			goto exit;
		n = n->next;
	}
	dev = NULL;
exit:
	spin_unlock(&pci_bus_lock);
	return dev;
}


/**
 *	device_resume_type - Resume some devices.
 *
 *	Resume devices of a specific type and their parents.
 *	Interrupts must be disabled when calling this.
 *
 *	Note that we only handle pci devices at the moment.
 *	We have no way that I can tell of getting the class
 *	of devices not on the pci bus.
 */
void device_resume_type(type)
{
	struct device * dev_dev;
	struct pci_dev * pci_dev = NULL;
	
	while ((dev = pci_find_class(PCI_BASE_CLASS_STORAGE, dev))) {
	}
}
#endif

EXPORT_SYMBOL(device_power_up);


