/*
 * suspend.c - Functions for moving devices between trees.
 *
 * Copyright (c) 2004 Nigel Cunningham
 *
 * This file is released under the GPLv2
 *
 */

#include <linux/device.h>
#include <linux/err.h>
#include "power.h"

/*
 *	device_merge_tree - Move an entire tree into another tree
 *	@source: The tree to be moved
 *	@dest  : The destination tree
 */

void device_merge_tree(	struct partial_device_tree * source,
			struct partial_device_tree * dest)
{
	down(&dpm_sem);
	list_splice_init(&source->dpm_active, &dest->dpm_active);
	list_splice_init(&source->dpm_off, &dest->dpm_off);
	list_splice_init(&source->dpm_off_irq, &dest->dpm_off_irq);
	up(&dpm_sem);
}
EXPORT_SYMBOL(device_merge_tree);

/*
 * 	device_switch_trees - Move a device and its ancestors to a new tree
 * 	@dev:	The lowest device to be moved.
 * 	@tree:	The destination tree.
 *
 * 	Note that siblings can be left in the original tree. This is because
 * 	we want to be able to keep part of a tree in one state while part is
 * 	in another.
 *
 * 	Since we iterate all the way back to the top, and may move entries
 * 	already in the destination tree, we will never violate the depth
 * 	first property of the destination tree.
 */

void device_switch_trees(struct device * dev, struct partial_device_tree * tree)
{
	down(&dpm_sem);
	while (dev) {
		list_del(&dev->power.entry);
		switch (dev->current_list) {
			case DEVICE_LIST_DPM_ACTIVE:
				list_add(&dev->power.entry, &tree->dpm_active);
				break;
			case DEVICE_LIST_DPM_OFF:
				list_add(&dev->power.entry, &tree->dpm_off);
				break;
			case DEVICE_LIST_DPM_OFF_IRQ:
				list_add(&dev->power.entry, &tree->dpm_off_irq);
				break;
		}
			
		dev = dev->parent;
	}
	up(&dpm_sem);
}

EXPORT_SYMBOL(device_switch_trees);

/*
 * 	create_device_tree - Create a new device tree
 */

struct partial_device_tree * device_create_tree(void)
{
	struct partial_device_tree * new_tree;

	new_tree = (struct partial_device_tree *) 
		kmalloc(sizeof(struct partial_device_tree),  GFP_ATOMIC);

	if (!IS_ERR(new_tree)) {
		INIT_LIST_HEAD(&new_tree->dpm_active);
		INIT_LIST_HEAD(&new_tree->dpm_off);
		INIT_LIST_HEAD(&new_tree->dpm_off_irq);
	}

	return new_tree;
}
EXPORT_SYMBOL(device_create_tree);

/*
 * 	device_destroy_tree - Destroy a dynamically created tree
 */

void device_destroy_tree(struct partial_device_tree * tree)
{
	BUG_ON(tree == &default_device_tree);
	
	BUG_ON(!list_empty(&tree->dpm_active));
	BUG_ON(!list_empty(&tree->dpm_off));
	BUG_ON(!list_empty(&tree->dpm_off_irq));

	kfree(tree);
}

EXPORT_SYMBOL(device_destroy_tree);
