/*
 * kernel/power/plugin.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It contains declarations for plugins. Plugins are additions to
 * suspend2 that provide facilities such as image compression or
 * encryption, backends for storage of the image and user interfaces.
 *
 */

/* This is the maximum size we store in the image header for a plugin name */
#define SUSPEND_MAX_PLUGIN_NAME_LENGTH 30

struct plugin_header {
	char name[SUSPEND_MAX_PLUGIN_NAME_LENGTH];
	int disabled;
	int type;
	int index;
	int data_length;
	unsigned long magic;
};

extern unsigned long memory_for_plugins(void);
extern int num_plugins;

#define FILTER_PLUGIN 1
#define WRITER_PLUGIN 2
#define UI_PLUGIN 3
#define MISC_PLUGIN 4 // Block writer, eg.
#define CHECKSUM_PLUGIN 5

#define SUSPEND_ASYNC 0
#define SUSPEND_SYNC  1

#define SUSPEND_COMMON_IO_OPS \
	/* Writing the image proper */ \
	int (*write_init) (int stream_number); \
	int (*write_chunk) (struct page * buffer_page); \
	int (*write_cleanup) (void); \
\
	/* Reading the image proper */ \
	int (*read_init) (int stream_number); \
	int (*read_chunk) (struct page * buffer_page, int sync); \
	int (*read_cleanup) (void); \
\
	/* Reset plugin if image exists but reading aborted */ \
	void (*noresume_reset) (void);

struct suspend_filter_ops {
	SUSPEND_COMMON_IO_OPS
	int (*expected_compression) (void);
	struct list_head filter_list;
};

struct suspend_writer_ops {
	
	SUSPEND_COMMON_IO_OPS

	/* Calls for allocating storage */

	long (*storage_available) (void); // Maximum size of image we can save
					  // (incl. space already allocated).
	
	unsigned long (*storage_allocated) (void);
					// Amount of storage already allocated
	int (*release_storage) (void);
	
	/* 
	 * Header space is allocated separately. Note that allocation
	 * of space for the header might result in allocated space 
	 * being stolen from the main pool if there is no unallocated
	 * space. We have to be able to allocate enough space for
	 * the header. We can eat memory to ensure there is enough
	 * for the main pool.
	 */
	long (*allocate_header_space) (unsigned long space_requested);
	int (*allocate_storage) (unsigned long space_requested);
	
	/* Read and write the metadata */	
	int (*write_header_init) (void);
	int (*write_header_chunk) (char * buffer_start, int buffer_size);
	int (*write_header_cleanup) (void);

	int (*read_header_init) (void);
	int (*read_header_chunk) (char * buffer_start, int buffer_size);
	int (*read_header_cleanup) (void);

	/* Prepare metadata to be saved (relativise/absolutise ranges) */
	int (*prepare_save_ranges) (void);
	int (*post_load_ranges) (void);
	
	/* Attempt to parse an image location */
	int (*parse_image_location) (char * buffer, int only_writer);

	/* Determine whether image exists that we can restore */
	int (*image_exists) (void);
	
	/* Mark the image as having tried to resume */
	void (*mark_resume_attempted) (void);

	/* Destroy image if one exists */
	int (*invalidate_image) (void);
	
	/* Wait on I/O */
	int (*wait_on_io) (int flush_all);

	struct list_head writer_list;
};

struct suspend_ui_ops {
	void (*early_boot_message_prep) (void);
	void (*prepare) (void);
	void (*message) (
		unsigned long type, unsigned long level,
		int normally_logged,
		const char *format, va_list args);
	void (*log_level_change) (void);
	unsigned long (*update_progress) (
		unsigned long value, unsigned long maximum,
		const char *fmt, va_list args);
	void (*cleanup) (void);
	int (*keypress) (unsigned int key);
	void (*post_kernel_restore_redraw) (void);

	struct list_head ui_list;
};

struct suspend_checksum_ops {
	void (*calculate_checksums) (void);
	void (*check_checksums) (void);
	void (*print_differences) (void);
	int (*allocate_pages) (void);
};

struct suspend_plugin_ops {
	/* Functions common to all plugins */
	int type;
	char * name;
	int disabled;
	struct list_head plugin_list;
	unsigned long (*memory_needed) (void);
	unsigned long (*storage_needed) (void);
	int (*print_debug_info) (char * buffer, int size);
	int (*save_config_info) (char * buffer);
	void (*load_config_info) (char * buffer, int len);
	
	/* Initialise & cleanup - general routines called
	 * at the start and end of a cycle. */
	int (*initialise) (void);
	void (*cleanup) (void);

	/* Set list of devices not to be suspended/resumed */
	void (*dpm_set_devices) (void);

	union {
		struct suspend_filter_ops filter;
		struct suspend_writer_ops writer;
		struct suspend_ui_ops ui;
		struct suspend_checksum_ops checksum;
	} ops;
};

extern struct suspend_plugin_ops * active_writer;
extern struct list_head suspend_filters, suspend_writers, suspend_plugins, suspend_ui;
extern struct suspend_plugin_ops * checksum_plugin;
extern void prepare_console_plugins(void);
extern void cleanup_console_plugins(void);
extern struct suspend_plugin_ops * find_plugin_given_name(char * name);
extern struct suspend_plugin_ops * get_next_filter(struct suspend_plugin_ops *);
extern int suspend_register_plugin(struct suspend_plugin_ops * plugin);
extern void suspend_move_plugin_tail(struct suspend_plugin_ops * plugin);

extern int initialise_suspend_plugins(void);
extern void cleanup_suspend_plugins(void);
extern unsigned long header_storage_for_plugins(void);
extern unsigned long memory_for_plugins(void);
extern int print_plugin_debug_info(char * buffer, int buffer_size);
extern int suspend_register_plugin(struct suspend_plugin_ops * plugin);
extern void suspend_unregister_plugin(struct suspend_plugin_ops * plugin);
extern int initialise_suspend_plugins(void);
extern void cleanup_suspend_plugins(void);
extern void suspend_post_restore_redraw(void);

static inline void suspend_checksum_calculate_checksums(void)
{
	if (checksum_plugin)
		checksum_plugin->ops.checksum.calculate_checksums();
}

static inline void suspend_checksum_print_differences(void)
{
	if (checksum_plugin)
		checksum_plugin->ops.checksum.print_differences();
}

static inline int suspend_allocate_checksum_pages(void)
{
	if (checksum_plugin)
		return checksum_plugin->ops.checksum.allocate_pages();
	else
		return 0;
}
