/*
 * kernel/power/process.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * Freeze_and_free contains the routines software suspend uses to freeze other
 * processes during the suspend cycle and to (if necessary) free up memory in
 * accordance with limitations on the image size.
 *
 * Ideally, the image saved to disk would be an atomic copy of the entire 
 * contents of all RAM and related hardware state. One of the first 
 * prerequisites for getting our approximation of this is stopping the activity
 * of other processes. We can't stop all other processes, however, since some 
 * are needed in doing the I/O to save the image. Freeze_and_free.c contains 
 * the routines that control suspension and resuming of these processes.
 * 
 * Under high I/O load, we need to be careful about the order in which we
 * freeze processes. If we freeze processes in the wrong order, we could
 * deadlock others. The freeze_order array this specifies the order in which
 * critical processes are frozen. All others are suspended after these have
 * entered the refrigerator.
 *
 * Another complicating factor is that freeing memory requires the processes
 * to not be frozen, but at the end of freeing memory, they need to be frozen
 * so that we can be sure we actually have eaten enough memory. This is why
 * freezing and freeing are in the one file. The freezer is not called from
 * the main logic, but indirectly, via the code for eating memory. The eat
 * memory logic is iterative, first freezing processes and checking the stats,
 * then (if necessary) unfreezing them and eating more memory until it looks 
 * like the criteria are met (at which point processes are frozen & stats
 * checked again).
 */

#define SUSPEND_FREEZER_C

#include <linux/module.h>
#include <linux/suspend.h>
#include <asm/tlbflush.h>

#include "suspend.h"

volatile struct suspend2_core_ops * suspend2_core_ops = NULL;
unsigned long suspend_action = 0;
unsigned long suspend_result = 0;
unsigned long suspend_debug_state = 0;
unsigned long software_suspend_state = ((1 << SUSPEND_DISABLED) | (1 << SUSPEND_BOOT_TIME) |
		(1 << SUSPEND_RESUME_NOT_DONE) | (1 << SUSPEND_IGNORE_LOGLEVEL));
unsigned int suspend_task = 0;

atomic_t __nosavedata suspend_cpu_counter = { 0 }; 

/* Timeouts when freezing */
#define FREEZER_TOTAL_TIMEOUT (5 * HZ)
#define FREEZER_CHECK_TIMEOUT (HZ / 10)

extern void suspend_relinquish_console(void);

/* ------------------------------------------------------------------------ */

/**
 * refrigerator - idle routine for frozen processes
 * @flag: unsigned long, non zero if signals to be flushed.
 *
 * A routine for kernel threads which should not do work during suspend
 * to enter and spin in until the process is finished.
 */

void refrigerator(unsigned long flag)
{
	unsigned long flags;
	long save;
	
	if (unlikely(current->flags & PF_NOFREEZE)) {
		current->flags &= ~PF_FREEZE;
		spin_lock_irqsave(&current->sighand->siglock, flags);
		recalc_sigpending();
		spin_unlock_irqrestore(&current->sighand->siglock, flags);
		return;
	}

	/* You need correct to work with real-time processes.
	   OTOH, this way one process may see (via /proc/) some other
	   process in stopped state (and thereby discovered we were
	   suspended. We probably do not care). 
	 */
	if ((flag) && (current->flags & PF_FREEZE)) {

		suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0,
			"\n%s (%d) refrigerated and sigpending recalculated.",
			current->comm, current->pid);
		spin_lock_irqsave(&current->sighand->siglock, flags);
		recalc_sigpending();
		spin_unlock_irqrestore(&current->sighand->siglock, flags);
	} else
		suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0,
			"\n%s (%d) refrigerated.",
			current->comm, current->pid);

	if (test_suspend_state(SUSPEND_FREEZER_ON)) {
		save = current->state;
		current->flags |= PF_FROZEN;
		while (current->flags & PF_FROZEN) {
			current->state = TASK_STOPPED;
			schedule();
			if (flag) {
				spin_lock_irqsave(
					&current->sighand->siglock, flags);
				recalc_sigpending();
				spin_unlock_irqrestore(
					&current->sighand->siglock, flags);
			}
		}
		current->state = save;
	} else
		suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0,
				"No longer freezing processes. Dropping out.\n");
	current->flags &= ~PF_FREEZE;
	spin_lock_irqsave(&current->sighand->siglock, flags);
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);
}


#ifdef CONFIG_SMP
static void __smp_pause(void * data)
{
	atomic_inc(&suspend_cpu_counter);
	while(test_suspend_state(SUSPEND_FREEZE_SMP)) {
		cpu_relax();
		barrier();
	}
	local_flush_tlb();
	atomic_dec(&suspend_cpu_counter);
}

void smp_pause(void)
{
	set_suspend_state(SUSPEND_FREEZE_SMP);
	smp_call_function(__smp_pause, NULL, 0, 0);

	while (atomic_read(&suspend_cpu_counter) < (num_online_cpus() - 1)) {
		cpu_relax();
		barrier();
	}
}

void smp_continue(void)
{
	clear_suspend_state(SUSPEND_FREEZE_SMP);
		
	while (atomic_read(&suspend_cpu_counter)) {
		cpu_relax();
		barrier();
	}
}

extern void __smp_suspend_lowlevel(void * info);

void smp_suspend(void)
{
	set_suspend_state(SUSPEND_FREEZE_SMP);
	smp_call_function(__smp_suspend_lowlevel, NULL, 0, 0);

	while (atomic_read(&suspend_cpu_counter) < (num_online_cpus() - 1)) {
		cpu_relax();
		barrier();
	}
}
#else
#define smp_pause() do { } while(0)
#define smp_continue() do { } while(0)
#define smp_suspend() do { } while(0)
#endif

/*
 * to_be_frozen
 *
 * Description:	Determine whether a process should be frozen yet.
 * Parameters:	struct task_struct *	The process to consider.
 * 		int			Which group of processes to consider.
 * Returns:	int			0 if don't freeze yet, otherwise do.
 */
static int to_be_frozen(struct task_struct * p, int type_being_frozen) {

	if ((p == current) ||
	    (p->flags & PF_NOFREEZE) ||
	    (p->state == TASK_ZOMBIE) ||
	    (p->state == TASK_TRACED) ||
	    (p->state == TASK_STOPPED))
		return 0;
	if ((!(p->mm)) && (type_being_frozen < 3))
		return 0;
	if ((p->flags & PF_SYNCTHREAD) && (type_being_frozen == 1))
		return 0;
	return 1;
}

/*
 * num_to_be_frozen
 *
 * Description:	Determine how many processes of our type are still to be
 * 		frozen. As a side effect, update the progress bar too.
 * Parameters:	int	Which type we are trying to freeze.
 * 		int	Whether we are displaying our progress.
 */
static int num_to_be_frozen(int type_being_frozen, int no_progress) {
	
	struct task_struct *p, *g;
	int todo_this_type = 0, total_todo = 0;
	int total_threads = 0;

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (to_be_frozen(p, type_being_frozen)) {
			todo_this_type++;
			total_todo++;
		} else if (to_be_frozen(p, 3))
			total_todo++;
		total_threads++;
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);

	if ((!no_progress) && (suspend2_core_ops)) {
		suspend2_core_ops->update_status(
				total_threads - total_todo,
				total_threads,
				"%d/%d", 
				total_threads - total_todo,
				total_threads);
	}
	return todo_this_type;
}

/*
 * freeze_threads
 *
 * Freeze a set of threads having particular attributes.
 *
 * Types:
 * 1: User threads not syncing.
 * 2: Remaining user threads.
 * 3: Kernel threads.
 */
extern void show_task(struct task_struct * p);

static int freeze_threads(int type, int no_progress)
{
	struct task_struct *p, *g;
	unsigned long start_time = jiffies;
	int result = 0, still_to_do;

	suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1,
		"\n STARTING TO FREEZE TYPE %d THREADS.\n",
		type);

	do {
		int numsignalled = 0;

		/* 
		 * Pause the other processors so we can safely
		 * change threads' flags
		 */
		smp_pause();

		if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
			smp_continue();
			return 1;
		}

		preempt_disable();

		local_irq_disable();

		read_lock(&tasklist_lock);

		/*
		 * Signal the processes.
		 *
		 * We signal them every time through. Otherwise pdflush -
		 * and maybe other processes - might never enter the
		 * fridge.
		 *
		 * NB: We're inside an SMP pause. Our printks are unsafe.
		 * They're only here for debugging.
		 *
		 */
		
		do_each_thread(g, p) {
			unsigned long flags;
			if (!to_be_frozen(p, type))
				continue;
			
			numsignalled++;
			suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0, 
				"\n   %s: pid %d",
				p->comm, p->pid);
			p->flags |= PF_FREEZE;
			spin_lock_irqsave(&p->sighand->siglock, flags);
			signal_wake_up(p, 0);
			spin_unlock_irqrestore(&p->sighand->siglock, flags);
		} while_each_thread(g, p);

		if (numsignalled)
			suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0,
				"\n Number of threads signalled this iteration is %d.\n",
				numsignalled);

		read_unlock(&tasklist_lock);

		/* 
		 * Let the processes run.
		 */		
		smp_continue();

		preempt_enable();

		local_irq_enable();
		
		/*
		 * Sleep.
		 */
		set_task_state(current, TASK_INTERRUPTIBLE);
		schedule_timeout(FREEZER_CHECK_TIMEOUT);

		still_to_do = num_to_be_frozen(type, no_progress);
	} while(still_to_do && (!TEST_RESULT_STATE(SUSPEND_ABORTED)) &&
		!time_after(jiffies, start_time + FREEZER_TOTAL_TIMEOUT));

	/*
	 * Did we time out? See if we failed to freeze processes as well.
	 *
	 */
	if ((time_after(jiffies, start_time + FREEZER_TOTAL_TIMEOUT)) && (still_to_do)) {
		read_lock(&tasklist_lock);
		do_each_thread(g, p) {
			if (!to_be_frozen(p, type)) 
				continue;
			
			if (!result) {
				printk(KERN_ERR name_suspend
					"Stopping tasks failed.\n");
				printk(KERN_ERR "Tasks that refused to be refrigerated"
					" and haven't since exited:\n");
				result = 1;
			}
			
			if (p->flags & PF_FREEZE) {
				printk(" - %s (#%d) signalled but "
					"didn't enter refrigerator.\n",
					p->comm, p->pid);
				show_task(p);
			} else
				printk(" - %s (#%d) wasn't "
					"signalled.\n",
					p->comm, p->pid);
		} while_each_thread(g, p);
		read_unlock(&tasklist_lock);
	} else
		suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1,
			"\n\nSuccessfully froze processes of type %d.\n",
			type);
	return result;
}

/*
 * freeze_processes - Freeze processes prior to saving an image of memory.
 * 
 * Return value: 0 = success, else # of processes that we failed to stop.
 */
extern asmlinkage long sys_sync(void);

/* Freeze_processes.
 * If the flag no_progress is non-zero, progress bars not be updated.
 * Debugging output is still printed.
 */
int freeze_processes(int no_progress)
{
	int showidlelist, result = 0, num_type[3];
	struct task_struct *p, *g;

	showidlelist = 1;

	num_type[0] = num_type[1] = num_type[2] = 0;

	set_suspend_state(SUSPEND_FREEZER_ON);

	suspend_result = 0;	/* Might be called from pm_disk or suspend -
				   ensure reset */

	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (p->mm) {
			if (p->flags & PF_SYNCTHREAD) {
				suspend_message(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
					"%s (%d) is a syncthread at entrance to "
					"fridge\n", p->comm, p->pid);
				num_type[1]++;
			} else
				num_type[2]++;
		} else {
			if (p->flags & PF_NOFREEZE)
				suspend_message(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0,
					"%s (%d) is NO_FREEZE.\n",
					p->comm, p->pid);
			else
				num_type[2]++;
		}
	} while_each_thread(g, p);
	read_unlock(&tasklist_lock);
	suspend_message(SUSPEND_FREEZER, SUSPEND_MEDIUM, 0, "\n");

	/* First, freeze all userspace,	non syncing threads. */
	if (freeze_threads(1, no_progress) || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto aborting;
	
	/* Now freeze processes that were syncing and are still running */
	if (freeze_threads(2, no_progress) || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto aborting;

	/* Now do our own sync, just in case one wasn't running already */
	if ((!no_progress) && (suspend2_core_ops))
		suspend2_core_ops->prepare_status(1, 1,
			"Freezing processes: Syncing remaining I/O.");

	sys_sync();

	set_suspend_state(SUSPEND_DISABLE_SYNCING);

	/* Freeze kernel threads */
	if (freeze_threads(3, no_progress) || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto aborting;

out:
	suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 1,
				"Left freezer loop.\n");

	/* No more syncing now! */
	suspend_task = current->pid;
	
	clear_suspend_state(SUSPEND_FREEZE_SMP);

	while (atomic_read(&suspend_cpu_counter)) {
		cpu_relax();
		barrier();
	}

	return result;
aborting:
	result = -1;
	goto out;
}

void thaw_processes(int which_threads)
{
	struct task_struct *p, *g;
	suspend_message(SUSPEND_FREEZER, SUSPEND_LOW, 1, "Thawing tasks\n");
	
	suspend_task = 0;
	if (which_threads != FREEZER_KERNEL_THREADS)
		clear_suspend_state(SUSPEND_FREEZER_ON);
	clear_suspend_state(SUSPEND_DISABLE_SYNCING);
	
	/* 
	 * Pause the other processors so we can safely
	 * change threads' flags
	 */

	smp_pause();

	preempt_disable();
	
	local_irq_disable();

	read_lock(&tasklist_lock);

	do_each_thread(g, p) {
		if (p->flags & PF_FROZEN) {
			if ((which_threads == FREEZER_KERNEL_THREADS) &&
				(p->mm))
				continue;
			suspend_message(SUSPEND_FREEZER, SUSPEND_VERBOSE, 0,
					"Waking %5d: %s.\n", p->pid, p->comm);
			p->flags &= ~PF_FROZEN;
			wake_up_process(p);
		}
	} while_each_thread(g, p);

	read_unlock(&tasklist_lock);

	smp_continue();
	
	preempt_enable();

	local_irq_enable();
}

EXPORT_SYMBOL(suspend_task);
EXPORT_SYMBOL(suspend_action);
EXPORT_SYMBOL(software_suspend_state);
EXPORT_SYMBOL(freeze_processes);
EXPORT_SYMBOL(thaw_processes);
#ifdef CONFIG_SMP
EXPORT_SYMBOL(smp_suspend);
EXPORT_SYMBOL(smp_continue);
#endif
EXPORT_SYMBOL(refrigerator);
