/*
 * kernel/power/range.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It contains declarations related to ranges. Ranges (otherwise
 * known as an extent, I'm told), is suspend's method of storing
 * all of the metadata for the image. See range.c for more info.
 *
 */

struct rangechain {
	struct range * first;
	struct range * last;
	int size; /* size of the range ie sum (max-min+1) */
	int allocs;
	int frees;
	int debug;
	int timesusedoptimisation;
	char * name;
	struct range * lastaccessed, *prevtolastaccessed, *prevtoprev;
};

/*
 * We rely on ranges not fitting evenly into a page.
 * The last four bytes are used to store the number
 * of the page, to make saving & reloading pages simpler.
 */
struct range {
	unsigned long minimum;
	unsigned long maximum;
	struct range * next;
};


#define RANGES_PER_PAGE (PAGE_SIZE / (sizeof(struct range)))
#define RANGEPAGELINK(x) ((unsigned long *) \
		((((unsigned long) x) & PAGE_MASK) + PAGE_SIZE - \
		 sizeof(unsigned long)))

#define range_for_each(rangechain, rangepointer, value) \
if ((rangechain)->first) \
	for ((rangepointer) = (rangechain)->first, (value) = \
			(rangepointer)->minimum; \
	     ((rangepointer) && ((rangepointer)->next || (value) <= \
				 (rangepointer)->maximum)); \
	     (((value) == (rangepointer)->maximum) ? \
		((rangepointer) = (rangepointer)->next, (value) = \
		 ((rangepointer) ? (rangepointer)->minimum : 0)) : \
			(value)++))

/*
 * When using compression and expected_compression > 0,
 * we allocate fewer swap entries, so GET_RANGE_NEXT can
 * validly run out of data to return.
 */
#define GET_RANGE_NEXT(currentrange, currentval) \
{ \
	if (currentrange) { \
		if ((currentval) == (currentrange)->maximum) { \
			if ((currentrange)->next) { \
				(currentrange) = (currentrange)->next; \
				(currentval) = (currentrange)->minimum; \
			} else { \
				(currentrange) = NULL; \
				(currentval) = 0; \
			} \
		} else \
			currentval++; \
	} \
}

extern int max_ranges_used;
extern int num_range_pages;
int add_to_range_chain(struct rangechain * chain, unsigned long value);
void put_range_chain(struct rangechain * chain);
void print_chain(int debuglevel, struct rangechain * chain, int printasswap);
int free_ranges(void);
int append_to_range_chain(int chain, unsigned long min, unsigned long max);
void relativise_ranges(void);
void relativise_chain(struct rangechain * chain);
void absolutise_ranges(void);
void absolutise_chain(struct rangechain * chain);
int get_rangepages_list(void);
void put_rangepages_list(void);
unsigned long * get_rangepages_list_entry(int index);
int relocate_rangepages(void);

extern struct range * first_range_page, * last_range_page;

#define RANGE_RELATIVE(x) (struct range *) ((((unsigned long) x) & \
			(PAGE_SIZE - 1)) | \
		((*RANGEPAGELINK(x) & (PAGE_SIZE - 1)) << PAGE_SHIFT))
#define RANGE_ABSOLUTE(entry) (struct range *) \
	((((unsigned long) (entry)) & (PAGE_SIZE - 1)) | \
	 (unsigned long) get_rangepages_list_entry(((unsigned long) (entry)) >> PAGE_SHIFT))

/* swap_entry_to_range_val & range_val_to_swap_entry: 
 * We are putting offset in the low bits so consecutive swap entries
 * make consecutive range values */
#define swap_entry_to_range_val(swp_entry) (swp_entry.val)
#define range_val_to_swap_entry(val) (swp_entry_t) { (val) }

