/*
 * kernel/power/suspend2.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It contains declarations used throughout swsusp and suspend2.
 *
 */
#ifndef KERNEL_POWER_SUSPEND_H
#define KERNEL_POWER_SUSPEND_H

#include <linux/delay.h>
#include "range.h"

/* ---------------------------- swsusp only ----------------------------- */

typedef struct pbe {
	unsigned long address;		/* address of the copy */
	unsigned long orig_address;	/* original address of page */
	swp_entry_t swap_address;	
	swp_entry_t dummy;		/* we need scratch space at 
					 * end of page (see link, diskpage)
					 */
} suspend_pagedir_t;

#define SUSPEND_PD_PAGES(x)     (((x)*sizeof(struct pbe))/PAGE_SIZE+1)
   
/* mm/page_alloc.c */
extern void drain_local_pages(void);

void save_processor_state(void);
void restore_processor_state(void);
struct saved_context;
void __save_processor_state(struct saved_context *ctxt);
void __restore_processor_state(struct saved_context *ctxt);

/* ---------------------------- Suspend2 -------------------------------- */

/* Page Backup Entry.
 *
 * This is an abstraction which contains the data for one
 * page of the image. (The data is really stored in ranges).
 */

struct pbe2 {
	struct page * origaddress;	/* Original address of page */
	struct page * address;		/* Address of copy of page */
	struct range * currentorigrange;
	struct range * currentdestrange;

	struct pagedir * pagedir;
};

/* Pagedir
 *
 * Contains the metadata for a set of pages saved in the image.
 */
struct pagedir {
	int pagedir_num;
	int pageset_size;
	int lastpageset_size;
	struct rangechain origranges;
	struct rangechain destranges;
	struct rangechain allocdranges;
};

/* Function for setting the chain names for a pagedir (used
 * for debugging */
void set_chain_names(struct pagedir * p);

#define pageset1_size (pagedir1.pageset_size)
#define pageset2_size (pagedir2.pageset_size)

/* Pagedir_nosave is pagedir1, loaded back in at the beginning
 * of resuming and relocated so we can do our atomic restoration
 * of the original kernel.
 * Pagedir1 is the metadata for pageset 1 pages. Ditto for pageset 2.
 */
extern suspend_pagedir_t *pagedir_nosave __nosavedata;
extern struct pagedir pagedir1, pagedir2;

/* Non-plugin data saved in our image header */
struct suspend_header {
	u32 version_code;
	unsigned long num_physpages;
	char machine[65];
	char version[65];
	int num_cpus;
	int page_size;
	unsigned long orig_mem_free;
	int num_range_pages;
	struct range * unused_ranges;
	int pageset_2_size;
	int param0;
	int param1;
	int param2;
	int param3;
	int param4;
	int progress0;
	int progress1;
	int progress2;
	int progress3;
	int io_time[2][2];
	
	/* Implementation specific variables */
#ifdef KERNEL_POWER_SWSUSP_C
	suspend_pagedir_t *suspend_pagedir;
	unsigned int num_pbes;
#else
	struct pagedir pagedir;
#endif
};

/* Suspend memory pool functions */
struct page * get_suspend_pool_pages(unsigned int gfp_mask, unsigned int order);
void free_suspend_pool_pages(struct page *page, unsigned int order);

extern void schedule_suspend_message(int message_number);
extern int suspend_min_free;

extern void suspend_restore_avenrun(void);
extern void suspend_save_avenrun(void);

extern unsigned long get_highstart_pfn(void);

#define SWAP_FILENAME_MAXLENGTH	32

extern int suspend_default_console_level;
extern int max_async_ios;
extern int image_size_limit;

struct pageset_sizes_result {
	int size1; /* Can't be unsigned - breaks MAX function */
	int size1low;
	int size2;
	int size2low;
	int needmorespace;
};

#define MB(x) ((x) >> (20 - PAGE_SHIFT))

extern int suspend_amount_grabbed;

/*
 * XXX: We try to keep some more pages free so that I/O operations succeed
 * without paging. Might this be more?
 */
#ifdef CONFIG_HIGHMEM
#define MIN_FREE_RAM (get_highstart_pfn() >> 7)
#else
#define MIN_FREE_RAM (max_mapnr >> 7)
#endif

extern void prepare_status(int printalways, int clearbar, const char *fmt, ...);
extern void abort_suspend(const char *fmt, ...);

extern int suspend_snprintf(char * buffer, int buffer_size,
		const char *fmt, ...);

/* ------ prepare_image.c ------ */
extern unsigned long get_grabbed_pages(int order);

/* ------ io.c ------ */
int suspend_early_boot_message(int can_erase_image, char *reason, ...);

/* ------ console.c ------ */
void check_shift_keys(int pause, char * message);
unsigned long update_status(unsigned long value, unsigned long maximum,
		const char *fmt, ...);

extern int expected_compression_ratio(void);

#define MAIN_STORAGE_NEEDED(USE_ECR) \
	((pageset1_size + pageset2_size) * \
	 (USE_ECR ? expected_compression_ratio() : 100) / 100)

#define HEADER_BYTES_NEEDED \
	((num_range_pages << PAGE_SHIFT) + \
	 sizeof(struct suspend_header) + \
	 sizeof(struct plugin_header) + \
	 (int) header_storage_for_plugins() + \
	 num_plugins * \
	 	(sizeof(struct plugin_header) + sizeof(int)))
	
#define HEADER_STORAGE_NEEDED ((HEADER_BYTES_NEEDED + (int) PAGE_SIZE - 1) >> PAGE_SHIFT)

#define STORAGE_NEEDED(USE_ECR) \
	(MAIN_STORAGE_NEEDED(USE_ECR) + HEADER_STORAGE_NEEDED)

#define RAM_TO_SUSPEND (1 + max((pageset1_size - pageset2_sizelow), 0) + \
		MIN_FREE_RAM + memory_for_plugins())

#ifndef KERNEL_POWER_SWSUSP_C
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
#define cond_show_pcp_lists() \
do { \
	if (TEST_DEBUG_STATE(SUSPEND_FREEZER)) \
		show_pcp_lists(); \
} while(0)

#define MDELAY(a) do { if (TEST_ACTION_STATE(SUSPEND_SLOW)) mdelay(a); } \
	while (0)
#define MAX_FREEMEM_SLOTS 24
enum {
	SUSPEND_FREE_BASE,
	SUSPEND_FREE_CONSOLE_ALLOC,
	SUSPEND_FREE_DRAIN_PCP,
	SUSPEND_FREE_IN_USE_MAP,
	SUSPEND_FREE_PS2_MAP,
	SUSPEND_FREE_CHECKSUM_MAP,
	SUSPEND_FREE_RELOAD_PAGES,
	SUSPEND_FREE_INIT_PLUGINS,
	SUSPEND_FREE_MEM_POOL,
	SUSPEND_FREE_FREEZER,
	SUSPEND_FREE_EAT_MEMORY,
	SUSPEND_FREE_SYNC,
	SUSPEND_FREE_GRABBED_MEMORY,
	SUSPEND_FREE_RANGE_PAGES,
	SUSPEND_FREE_EXTRA_PD1,
	SUSPEND_FREE_WRITER_STORAGE,
	SUSPEND_FREE_HEADER_STORAGE,
	SUSPEND_FREE_CHECKSUM_PAGES,
	SUSPEND_FREE_KSTAT,
	SUSPEND_FREE_DEBUG_INFO,
	SUSPEND_FREE_INVALIDATE_IMAGE,
	SUSPEND_FREE_IO,
	SUSPEND_FREE_IO_INFO,
	SUSPEND_FREE_START_ONE
};
extern void suspend_store_free_mem(int slot, int side);
extern int suspend_free_mem_values[MAX_FREEMEM_SLOTS][2];
#else
#define suspend_store_free_mem(a, b) do { } while(0)
#define MDELAY(a) do { } while (0)
#define cond_show_pcp_lists() do { } while(0)
#endif
#endif /* Not swsusp */

extern int expected_compression_ratio(void);
int print_module_list_to_buffer(char * buffer, int size);

extern unsigned int nr_suspends;
extern char resume2_file[];

extern int suspend_wait_for_keypress(void);

#ifdef CONFIG_SMP
extern void smp_suspend(void);
extern void smp_continue(void);
#else
#define smp_suspend() do { } while(0)
#define smp_continue() do { } while(0)
#endif

/* For user interface */
#include <linux/syscalls.h>
extern asmlinkage ssize_t sys_write(unsigned int fd, const char __user * buf, 
	size_t count);

#ifdef CONFIG_BOOTSPLASH
#include <linux/console.h>
#include <linux/fb.h>
#include "../../drivers/video/console/fbcon.h"
static inline struct splash_data * get_splash_data(int consolenr)
{
	BUG_ON(consolenr >= MAX_NR_CONSOLES);

	if (vc_cons[consolenr].d)
		return vc_cons[consolenr].d->vc_splash_data;
	
	return NULL;
}
#endif

extern asmlinkage ssize_t sys_write(unsigned int fd, const char __user * buf, size_t count);

extern struct pm_ops * pm_ops;
extern dev_t name_to_dev_t(char *line);
extern char _text[], _etext[], _edata[], __bss_start[], _end[];
extern void signal_wake_up(struct task_struct *t, int resume);

extern struct partial_device_tree * suspend_device_tree;

/* Returns whether it was already in the requested state */
extern int suspend_map_kernel_page(struct page * page, int enable);

#ifdef CONFIG_DEBUG_PAGEALLOC
extern void suspend_map_atomic_copy_pages(void);
extern void suspend_unmap_atomic_copy_pages(void);
#else
#define suspend_map_atomic_copy_pages() do { } while(0)
#define suspend_unmap_atomic_copy_pages() do { } while(0)
#endif

#endif
