/*
 * kernel/power/suspend2-builtin.c
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It contains the functions for suspend2 that are built into the kernel even if
 * suspend is configured as modules.
 */

#include <linux/suspend.h>
#include <linux/module.h>
#include <linux/reboot.h>
#include <asm/highmem.h>
#include <asm/uaccess.h>

#include "suspend.h"
/* 
 *---------------------  Variables ---------------------------
 * 
 * The following are used by the arch specific low level routines 
 * and only needed if suspend2 is compiled in. Other variables,
 * used by the freezer even if suspend2 is not compiled in are
 * found in process.c
 */
volatile int suspend_io_time[2][2];
struct pagedir __nosavedata pagedir_resume;
struct pagedir pagedir1 = { 1, 0, 0}, pagedir2 = {2, 0, 0};
static unsigned long avenrun_save[3];

char suspend_print_buf[1024];
EXPORT_SYMBOL(suspend_print_buf);

/* Suspend2 variables used by built-in routines. */
unsigned int nr_suspends = 0;
int suspend_act_used = 0;
int suspend_lvl_used = 0;
int suspend_dbg_used = 0;
int suspend_default_console_level = 0;

/* 
 * For resume2= kernel option. It's pointless to compile
 * suspend2 without any writers, but compilation shouldn't
 * fail if you do.
 */
#ifdef CONFIG_SOFTWARE_SUSPEND_DEFAULT_RESUME2
char resume2_file[256] = CONFIG_SOFTWARE_SUSPEND_DEFAULT_RESUME2;
#else
char resume2_file[256];
#endif

void (* exclusive_handler) (int) = NULL;

/* --------------- Basic user interface functions --------------- 
 * 
 * These need to be available even if none of the core is
 * loaded.
 */

DECLARE_WAIT_QUEUE_HEAD(suspend_wait_for_key);
static int last_key;

int suspend_wait_for_keypress(void)
{
	interruptible_sleep_on(&suspend_wait_for_key);
	return last_key;
}

/* 
 * Basic keypress handler for suspend. This is extensible
 * via the user interface modules.
 */

/* For simplicity, we convert keyboard key codes to ascii,
 * except in the case of function keys, which are mapped
 * to 1-12. We can then use the same case statement for
 * serial keyboards (and from a serial keyboard, you can
 * press Control-A..L to toggle sections.
 */
static unsigned int kbd_keytable[] = {
	  0,  27,  49,  50,  51,  52,  53,  54,  55,  56,
	 57,  48,   0,   0,   0,   0,   0,   0,   0, 114,
	  0,   0,   0,   0,   0, 112,   0,   0,   0,   0,
	  0, 115,   0,   0,   0,   0,   0,   0, 108,   0,
	  0, 122,   0,   0,   0,   0,  99,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,  32,   0,   1,
	  2,   3,   4,   5,   6,   7,   8,   9,  10,   0,
	  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,  11,  12,   0,
};

/*
 * keycode_to_action
 *
 * Convert a keycode (serial or keyboard) into our
 * internal code (ascii, except for function keys).
 */
static unsigned int keycode_to_action(unsigned int keycode, int source)
{
	if (source == SUSPEND_KEY_SERIAL) {
		if (keycode > 64)
			return (keycode | 32);
		else
			return keycode;
	}

	/* Local keyboard - use table above */
	if (keycode > sizeof(kbd_keytable))
		return 0;

	return kbd_keytable[keycode];
}

/* get_keyboard_exclusive
 *
 * Give a plugin exclusive access to the keyboard
 * if it's not already claimed. Used for an encryption
 * plugin to get the passphrase, for example.
 */

int suspend_get_keyboard_exclusive(void (* handler) (int))
{
	if (exclusive_handler)
		return -EBUSY;

	exclusive_handler = handler;

	return 0;
}

/* release_keyboard_exclusive
 *
 * Release the exclusive access to the keyboard.
 */

void suspend_release_keyboard_exclusive(void)
{
	BUG_ON(!exclusive_handler);

	exclusive_handler = NULL;
}

/*
 * suspend_handle_keypress
 *
 * This is the basic routine for handling keystrokes.
 * If it doesn't know what to do with a keypress, it
 * passes it on to the plugins.
 */
void suspend_handle_keypress(unsigned int keycode, int source)
{
	keycode = keycode_to_action(keycode, source);

	if (test_suspend_state(SUSPEND_SANITY_CHECK_PROMPT)) {
		if (keycode == 32)
			wake_up_interruptible(&suspend_wait_for_key);
		else if (keycode == 99) {
			set_suspend_state(SUSPEND_CONTINUE_REQ);
			wake_up_interruptible(&suspend_wait_for_key);
		}
		return;
	}

	/* Do we have a plugin grabbing all the keys? */
	if (exclusive_handler) {
		exclusive_handler(keycode);
		return;
	}

	last_key = keycode;

	/* 
	 * If the message was handled or is the space bar, we
	 * wake our completion handler.
	 */
	if ((suspend2_core_ops->keypress(keycode)) ||
		(keycode == 32))
		wake_up_interruptible(&suspend_wait_for_key);
}

/* suspend_early_boot_message()
 * Description:	Handle errors early in the process of booting.
 * 		The user may press C to continue booting, perhaps
 * 		invalidating the image,  or space to reboot. 
 * 		This works from either the serial console or normally 
 * 		attached keyboard.
 *
 * 		Note that we come in here from init, while the kernel is
 * 		locked. If we want to get events from the serial console,
 * 		we need to temporarily unlock the kernel.
 *
 * Arguments:	Char *. Pointer to a string explaining why we're moaning.
 */

#define say(message, a...) printk(KERN_EMERG message, ##a)

int suspend_early_boot_message(int can_erase_image, char *warning_reason, ...)
{
	unsigned long orig_state = get_suspend_state(), continue_req;
	int orig_loglevel = console_loglevel;
	va_list args;
	int printed_len;

	set_suspend_state(SUSPEND_RUNNING);

#ifdef CONFIG_BOOTSPLASH
/* 
 * So we can make any error visible if necessary. The core might
 * not be loaded and the bootsplash module might not be loaded
 * or even have been compiled.
 */
	{
		extern int splash_verbose(void);
		splash_verbose();
	}
#endif

#if defined(CONFIG_VT) || defined(CONFIG_SERIAL_CONSOLE)
	console_loglevel = 7;

	if (suspend2_core_ops)
		suspend2_core_ops->early_boot_plugins();

	say("=== Software Suspend ===\n\n");
	if (warning_reason) {
		va_start(args, warning_reason);
		printed_len = vsnprintf(suspend_print_buf, 
				sizeof(suspend_print_buf), 
				warning_reason,
				args);
		va_end(args);
		say("BIG FAT WARNING!! %s\n\n", suspend_print_buf);
		if (can_erase_image) {
			say("If you want to use the current suspend image, reboot and try\n");
			say("again with the same kernel that you suspended from. If you want\n");
			say("to forget that image, continue and the image will be erased.\n");
		} else {
			say("If you continue booting, note that any image WILL NOT BE REMOVED.\n");
			say("Suspend is unable to do so because the appropriate modules aren't\n");
			say("loaded. You should manually remove the image to avoid any\n");
			say("possibility of corrupting your filesystem(s) later.\n");
		}
		say("Press SPACE to reboot or C to continue booting with this kernel\n");
	} else {
		say("BIG FAT WARNING!!\n\n");
		say("You have tried to resume from this image before.\n");
		say("If it failed once, it will probably fail again.\n");
		say("Would you like to remove the image and boot normally?\n");
		say("This will be equivalent to entering noresume2 on the\n");
		say("kernel command line.\n\n");
		say("Press SPACE to remove the image or C to continue resuming.\n");
	}
	
	set_suspend_state(SUSPEND_SANITY_CHECK_PROMPT);

	interruptible_sleep_on(&suspend_wait_for_key);

	continue_req = test_suspend_state(SUSPEND_CONTINUE_REQ);
	
	if ((warning_reason) && (!continue_req))
		machine_restart(NULL);
	
	restore_suspend_state(orig_state);
	if (continue_req)
		set_suspend_state(SUSPEND_CONTINUE_REQ);

	console_loglevel = orig_loglevel;
#endif // CONFIG_VT or CONFIG_SERIAL_CONSOLE
	return -EPERM;
}
#undef say

/* --------------- Registration of the core code -----------------------------
 *
 * We don't need to do anything more. The writers determine
 * whether suspending is disabled and they're not loaded yet.
 */
int suspend2_register_core(struct suspend2_core_ops * ops_pointer)
{
	if (suspend2_core_ops)
		return -EBUSY;

	suspend2_core_ops = ops_pointer;
	return 0;
}

void suspend2_unregister_core(void)
{
	suspend2_core_ops = NULL;
}

/* ------------ Functions for kickstarting a suspend or resume ----------- */

static int can_suspend(void)
{
	if (test_suspend_state(SUSPEND_RUNNING)) {
		printk(name_suspend "Software suspend is already running.\n");
		return 0;
	}

	if (!suspend2_core_ops) {
		printk(name_suspend "Software suspend is disabled.\n"
			"You do not appear to have inserted the core module.\n");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		return 0;
	}

	if (test_suspend_state(SUSPEND_DISABLED)) {
		printk(name_suspend "Software suspend is disabled.\n"
			"This may be because you haven't put something along the "
			"lines of\n\nresume2=swap:/dev/hda1\n\n"
			"in lilo.conf or equivalent. (Where /dev/hda1 is your "
			"swap partition).\n");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		return 0;
	}
	
	return 1;
}

/*
 * Check if we have an image and if so try to resume.
 */

void software_suspend_try_resume(void)
{
	mm_segment_t oldfs;
	oldfs = get_fs(); set_fs(KERNEL_DS);

	clear_suspend_state(SUSPEND_RESUME_NOT_DONE);

	if (!suspend2_core_ops) {
		/*
		 * We can only get here at boot time. It is really dangerous
		 * to suspend, boot without resuming and then boot with
		 * resuming. Since the core (and writers) isn't loaded, we
		 * can't fix this. We can however print a big fat warning
		 * and give the user the option of rebooting.
		 *
		 * We don't do this if no resume2= parameter was specified.
		 */

		if (resume2_file[0])
			suspend_early_boot_message(0,
			 "Can't check whether to resume. Suspend's core module isn't loaded.");
		goto out;
	}
	suspend2_core_ops->do_resume();
out:
	set_fs(oldfs);
	clear_suspend_state(SUSPEND_IGNORE_LOGLEVEL);
	return;
}

/*
 * suspend_try_suspend
 * Functionality   : First level of code for software suspend invocations.
 * 		     Performs the basic checking as to whether suspend is
 * 		     enabled before invoking the high level routine.
 * Called From     : 
 */
void suspend_try_suspend(void)
{
	mm_segment_t oldfs;

	if (!can_suspend())
		return;
	
	oldfs = get_fs(); set_fs(KERNEL_DS);
	suspend2_core_ops->do_suspend();
	set_fs(oldfs);
}

/* -------------------  Commandline Parameter Handling -----------------
 *
 * Resume setup: obtain the storage device.
 */

static int __init resume_setup(char *str)
{
	if (str == NULL)
		return 1;
	
	strncpy(resume2_file, str, 255);
	return 0;
}

/*
 * Allow the user to set the action parameter from lilo, prior to resuming.
 */
static int __init suspend_act_setup(char *str)
{
	if(str)
		suspend_action=simple_strtol(str,NULL,0);
	suspend_act_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug parameter from lilo, prior to resuming.
 */
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
static int __init suspend_dbg_setup(char *str)
{
	if(str)
		suspend_debug_state=simple_strtol(str,NULL,0);
	suspend_dbg_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug level parameter from lilo, prior to
 * resuming.
 */
static int __init suspend_lvl_setup(char *str)
{
	if(str)
		console_loglevel =
		suspend_default_console_level = 
			simple_strtol(str,NULL,0);
	suspend_lvl_used = 1;
	clear_suspend_state(SUSPEND_IGNORE_LOGLEVEL);
	return 0;
}
#endif

/*
 * Allow the user to specify that we should ignore any image found and
 * invalidate the image if necesssary. This is equivalent to running
 * the task queue and a sync and then turning off the power. The same
 * precautions should be taken: fsck if you're not journalled.
 */
static int __init noresume_setup(char *str)
{
	set_suspend_state(SUSPEND_NORESUME_SPECIFIED);
	/* Message printed later */
	return 0;
}

/* In leiu of exporting variables, some get_ functions for suspend2 */
unsigned long get_highstart_pfn(void)
{
	return highstart_pfn;
}

/*
 * Running suspend makes for a very high load average. I'm told that
 * sendmail and crond check the load average, so in order for them
 * not to be unnecessarily affected by the operation of suspend, we
 * store the avenrun values prior to suspending and restore them
 * at the end of the resume cycle. Thus, the operation of suspend
 * should be invisible to them. Thanks to Marcus Gaugusch and Bernard
 * Blackham for noticing the problem and suggesting the solution.
 */
void suspend_save_avenrun(void)
{
	int i;

	for (i = 0; i < 3; i++)
		avenrun_save[i] = avenrun[i];
}

void suspend_restore_avenrun(void)
{
	int i;

	for (i = 0; i < 3; i++)
		avenrun[i] = avenrun_save[i];
}

__setup("resume2=", resume_setup);
__setup("suspend_act=", suspend_act_setup);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
__setup("suspend_dbg=", suspend_dbg_setup);
__setup("suspend_lvl=", suspend_lvl_setup);
#endif
__setup("noresume2", noresume_setup);

EXPORT_SYMBOL(get_highstart_pfn);
EXPORT_SYMBOL(suspend_save_avenrun);
EXPORT_SYMBOL(suspend_restore_avenrun);

EXPORT_SYMBOL(nr_suspends);
EXPORT_SYMBOL(pagedir1);
EXPORT_SYMBOL(pagedir2);
EXPORT_SYMBOL(suspend2_register_core);
EXPORT_SYMBOL(suspend2_unregister_core);
EXPORT_SYMBOL(resume2_file);
EXPORT_SYMBOL(suspend_act_used);
EXPORT_SYMBOL(suspend_lvl_used);
EXPORT_SYMBOL(suspend_dbg_used);
EXPORT_SYMBOL(suspend_try_suspend);
EXPORT_SYMBOL(suspend_debug_state);
EXPORT_SYMBOL(suspend_result);

/* Symnols exported for Suspend plugins */
EXPORT_SYMBOL(suspend_default_console_level);
EXPORT_SYMBOL(pagedir_resume);
EXPORT_SYMBOL(suspend_io_time);
EXPORT_SYMBOL(suspend2_core_ops);
EXPORT_SYMBOL(suspend_early_boot_message);
EXPORT_SYMBOL(suspend_wait_for_keypress);
