/*
 * kernel/power/suspend_dm.c
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * This file contains support for interfacing with the device mapper
 * to allocate memory for its work.
 */

#include <linux/suspend.h>
#include <linux/module.h>

/* For calculating how much memory it needs */
#include "../../drivers/md/dm-io.h"

#include "suspend.h"
#include "plugins.h"
#include "proc.h"

static struct suspend_plugin_ops suspend_dm_ops;
static int io_get_result;

/* ---- Exported functions ---- */

/* suspend_dm_init()
 *
 * Description:	Allocate buffers for device mapper use.
 * Returns:	Zero on success, -ENOMEM if unable to vmalloc.
 */

static int suspend_dm_init(void)
{
	io_get_result = dm_io_get(max_async_ios);
	return io_get_result;
}

/* suspend_dm_cleanup()
 *
 * Description: Tell DM to release the memory we allocated.
 * Returns:	Zero. Always works!
 */

static void suspend_dm_cleanup(void)
{
	if (!io_get_result)
		dm_io_put(max_async_ios);
}

/* suspend_dm_save_config_info
 *
 * Description:	Save informaton needed when reloading the image at resume time.
 * Arguments:	Buffer:		Pointer to a buffer of size PAGE_SIZE.
 * Returns:	Number of bytes used for saving our data.
 */

static int suspend_dm_save_config_info(char * buffer)
{
	return 0;
}

/* suspend_dm_load_config_info
 *
 * Description:	Reload information needed for decompressing the image at 
 * 		resume time.
 * Arguments:	Buffer:		Pointer to the start of the data.
 *		Size:		Number of bytes that were saved.
 */

static void suspend_dm_load_config_info(char * buffer, int size)
{
	BUG_ON(size);
	return;
}

/*
 * data for our proc entries.
 */

static struct suspend_proc_data disable_dm_support_proc_data = {
	.filename			= "disable_device_mapper_support",
	.permissions			= PROC_RW,
	.type				= SUSPEND_PROC_DATA_INTEGER,
	.data = {
		.integer = {
			.variable	= &suspend_dm_ops.disabled,
			.minimum	= 0,
			.maximum	= 1,
		}
	}
};

/*
 * Ops structure.
 */

static struct suspend_plugin_ops suspend_dm_ops = {
	.type			= MISC_PLUGIN,
	.name			= "Device Mapper Support",
	.initialise		= suspend_dm_init,
	.cleanup		= suspend_dm_cleanup,
	.save_config_info	= suspend_dm_save_config_info,
	.load_config_info	= suspend_dm_load_config_info,
};

/* ---- Registration ---- */

static __init int suspend_dm_load(void)
{
	int result;

	if (!(result = suspend_register_plugin(&suspend_dm_ops))) {
		printk("Software Suspend Device Mapper support registering.\n");
		suspend_register_procfile(&disable_dm_support_proc_data);
	}
	return result;
}

#ifdef MODULE
static __exit void suspend_dm_unload(void)
{
	printk("Software Suspend Device Mapper support unloading.\n");
	suspend_unregister_procfile(&disable_dm_support_proc_data);
	suspend_unregister_plugin(&suspend_dm_ops);
}


module_init(suspend_dm_load);
module_exit(suspend_dm_unload);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nigel Cunningham");
MODULE_DESCRIPTION("Device Mapper support for Suspend2");
#else
late_initcall(suspend_dm_load);
#endif
